"""
objects.py
--------------

Deal with objects which hold visual properties, like
ColorVisuals and TextureVisuals.
"""
import numpy as np

from .color import ColorVisuals
from ..util import log


def create_visual(**kwargs):
    """
    Create Visuals object from keyword arguments.

    Parameters
    ----------
    face_colors   :   (n,3|4) uint8, colors
    vertex_colors : (n,3|4) uint8, colors
    mesh:          Trimesh object

    Returns
    ----------
    visuals: ColorVisuals object.
    """
    return ColorVisuals(**kwargs)


def concatenate(visuals, *args):
    """
    Concatenate multiple visual objects.

    Parameters
    ----------
    visuals: ColorVisuals object, or list of same
    *args:  ColorVisuals object, or list of same

    Returns
    ----------
    concat: ColorVisuals object
    """
    # get a flat list of ColorVisuals objects
    if len(args) > 0:
        visuals = np.append(visuals, args)
    else:
        visuals = np.array(visuals)

    try:
        # get the mode of the first visual
        mode = visuals[0].kind
        if mode == 'face':
            colors = np.vstack([
                v.face_colors for v in visuals])
            return ColorVisuals(face_colors=colors)
        elif mode == 'vertex':
            colors = np.vstack([
                v.vertex_colors for v in visuals])
            return ColorVisuals(vertex_colors=colors)
    except BaseException:
        log.warning('failed to concatenate visuals!',
                    exc_info=True)

    return ColorVisuals()



"""
schemas.py
-------------

Tools for dealing with schemas, particularly JSONschema
"""
import json


def resolve_json(text, resolver, recursive=True, fill_empty='{}'):
    """
    Given a JSON Schema containing `$ref` keys, replace all
    referenced URI values with their values using trimesh
    Resolver objects.

    Parameters
    ---------------
    text : str
      JSON text including `$ref` to other files
    resolver : trimesh.visual.resolver.Resolver
      Resolver to fetch referenced assets
    recursive : bool
      If True, resolve references in referenced files
    fill_empty : str
      What to replace empty references with

    Returns
    ----------
    result : str
      JSON text with references filled in
    """
    if '$ref' not in text:
        return text
    # append strings to list then concatenate once
    result = []
    # set the current value to the input text
    current = text
    # loop with a for to cap iterations
    for i in range(len(text)):
        # find the referenced key in the text
        idx = current.find('$ref')
        # if the string wasn't found find will return -1
        if idx < 0:
            break
        # find the first opening bracket before the ref key
        first = current.rfind('{', 0, idx)
        # find the first closing bracket after the ref key
        last = current.find('}', idx)

        # extract the URI of the reference
        name = json.loads(current[first:last + 1])['$ref']

        # get the bytes from the resolver
        data = resolver.archive[name].read().decode('utf-8')

        # include all text before the first reference
        result.append(current[:first])

        if recursive:
            result.append(
                resolve_json(
                    data, resolver=resolver, recursive=True))
        else:
            # include the loaded data
            result.append(data)

        # if we had a reference to an empty file fill it
        if len(result[-1].strip()) == 0:
            result.append(fill_empty)

        # for the next loop only look at the next chunk
        current = current[last + 1:]

    # any values after the last reference
    result.append(current)

    # append results into single string
    appended = ' '.join(result)

    # assert we got rid of all the references
    assert '$ref' not in appended

    return appended



\
import os

# find the current absolute path to this directory
_pwd = os.path.expanduser(os.path.abspath(os.path.dirname(__file__)))

# once resources are loaded cache them
_cache = {}


def get(name, decode=True):
    """
    Get a resource from the trimesh/resources folder.

    Parameters
    -------------
    name : str
      File path relative to `trimesh/resources`
    decode : bool
      Whether or not to decode result as UTF-8

    Returns
    -------------
    resource : str or bytes
      File data
    """
    # key by name and decode
    cache_key = (name, bool(decode))
    if cache_key in _cache:
        # return cached resource
        return _cache[cache_key]

    # get the resource using relative names
    with open(os.path.join(_pwd, name), 'rb') as f:
        resource = f.read()

    # make sure we return it as a string if asked
    if decode and hasattr(resource, 'decode'):
        resource = resource.decode('utf-8')
    # store for later access
    _cache[cache_key] = resource

    return resource



"""
https://github.com/mikedh/trimesh
------------------------------------

Trimesh is a pure Python (2.7- 3.3+) library for loading and using triangular
meshes with an emphasis on watertight meshes. The goal of the library is to
provide a fully featured Trimesh object which allows for easy manipulation
and analysis, in the style of the Polygon object in the Shapely library.
"""

# current version
from .version import __version__

# geometry objects
from .base import Trimesh
from .points import PointCloud
from .scene.scene import Scene

# utility functions
from .util import unitize
from .transformations import transform_points

# general numeric tolerances
from .constants import tol

# loader functions
from .exchange.load import (load,
                            load_mesh,
                            load_path,
                            load_remote,
                            available_formats)

# avoid a circular import in trimesh.base
from . import voxel
from . import primitives

# explicitly list imports in __all__
# as otherwise flake8 gets mad
__all__ = [__version__,
           'Trimesh',
           'PointCloud',
           'Scene',
           'voxel',
           'unitize',
           'tol',
           'load',
           'load_mesh',
           'load_path',
           'load_remote',
           'primitives',
           'transform_points',
           'available_formats']
